#include "ISocket.h"
#include "ComplementsPaires.h"
#include "FamillesAdresse.h"
#include "FamillesProtocole.h"
#include <winsock2.h>
#include <algorithm>
#include <utility>
using std::pair;

namespace SocketsFlux
{
   class SocketWin32
      : public ISocket
   {
      SOCKET socket_;
      int famille_, protocole_;
   public:
      SocketWin32(const FamilleAdresse fam, const FamilleProtocole proto)
         : socket_(INVALID_SOCKET),
           famille_(GestionnaireFamilles::get().trouver(fam)),
           protocole_(GestionnaireProtocoles::get().trouver(proto))
      {
         socket_ = socket(famille_, SOCK_STREAM, protocole_);
         if (socket_ == INVALID_SOCKET)
            throw ErreurCreation();
      }
      SocketWin32(const SOCKET sck, const int fam, const int proto)
         : socket_(sck), famille_(fam), protocole_ (proto)
      {
      }
      pair<EtatConnexion, erreur_type>
         connecter(const Adresse &a, const Port &p)
      {
         EtatConnexion e = ConnexionEnCours;
         erreur_type err = 0;
         //
         // Mcanique de l'appel  ::connect()
         //
         sockaddr_in adresse = { 0 };
         adresse.sin_family  = famille_;
         adresse.sin_addr.s_addr = inet_addr(a.valeur().c_str());
         adresse.sin_port = p.valeur(Normalisee());
         //
         if (connect(socket_, reinterpret_cast<const sockaddr*> (&adresse),
                     sizeof (sockaddr)) == SOCKET_ERROR)
         {
            switch (err = WSAGetLastError())
            {
            case WSAEALREADY:
            case WSAEWOULDBLOCK:
               e = ConnexionEnCours;
               break;
            case WSAEISCONN:
               e = ConnexionOk;
               break;
            default:
               e = ConnexionErreur;
            }
         }
         else
            e = ConnexionOk;
         return pair<EtatConnexion,erreur_type>(e,err);
      }
      bool fermer() throw()
      {
         if (socket_ == INVALID_SOCKET)
            return false;
         shutdown(socket_, SD_BOTH);
         closesocket(socket_);
         return true;
      }
      bool lier(const Port &p)
      {
         const int TAILLE_FILE_ATTENTE = 16;
         sockaddr_in sa = { 0 };
         sa.sin_family  = famille_;
         sa.sin_port    = p.valeur(Normalisee());
         return bind(socket_, reinterpret_cast<const sockaddr*>(&sa),
                     sizeof(sockaddr_in)) != SOCKET_ERROR &&
                listen(socket_, TAILLE_FILE_ATTENTE) != SOCKET_ERROR;
      }
      int envoyerPrimitif(const char *seq, const int n)
      {
         int n_env = send (socket_, seq, n, 0);
         if (n_env == SOCKET_ERROR)
            throw pair<
               FermetureHomologue,erreur_type
            >(FermetureHomologue(false), WSAGetLastError());
         return n_env;
      }
      int recevoirPrimitif(char *dest, const size_t n)
      {
         int n_rec = recv(socket_, dest, static_cast<int>(n), 0);
         if (!n_rec)
         {
            erreur_type err = WSAGetLastError();
            if (err != WSAEWOULDBLOCK)
               throw pair<
                  FermetureHomologue,erreur_type
               >(FermetureHomologue(),err);
         }
         else if (n_rec == SOCKET_ERROR)
            throw pair<
               FermetureHomologue,erreur_type
            >(FermetureHomologue(false),WSAGetLastError());
         return n_rec;
      }
      void rendreBloquant()
      {
         unsigned long mode = 0;
         ioctlsocket(socket_, FIONBIO, &mode);
      }
      void rendreNonBloquant()
      {
         unsigned long mode = 1; // tout non zro fait
         ioctlsocket(socket_, FIONBIO, &mode);
      }
      // Mcanique d'acceptation
      pair <EtatAcceptation,ISocket*> accepter()
      {
         sockaddr_in homologue = { 0 };
         int tailleHomologue= sizeof(sockaddr_in);
         SOCKET sck = accept(socket_, reinterpret_cast<sockaddr *>(&homologue),
                             &tailleHomologue);
         EtatAcceptation e = AcceptationOk;
         ISocket *s = 0;
         if (sck == INVALID_SOCKET)
         {
            erreur_type err = WSAGetLastError();
            switch (err)
            {
            case WSAEWOULDBLOCK:
               e = AcceptationPersonne;
               break;
            default:
               e = AcceptationErreur;
            }
         }
         else
            s = new SocketWin32(sck, famille_, protocole_);
         return pair<EtatAcceptation,ISocket*>(e, s);
      }
   };
   ISocket *ISocket::Creer(const FamilleAdresse fam, const FamilleProtocole proto)
      { return new SocketWin32(fam, proto); }
}
